"""Copyright 2024 PythonistaGuild

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from collections.abc import Awaitable, Callable
from typing import Literal, TypeAlias, TypedDict

from starlette.requests import Request
from starlette.responses import Response


BucketCallable: TypeAlias = Callable[[Request], Awaitable[str | None]]
ExemptCallable: TypeAlias = Callable[[Request], Awaitable[bool]]

BucketType: TypeAlias = Literal["ip"] | BucketCallable
ResponseCallback: TypeAlias = Callable[[Request, float], Awaitable[Response]]


class RateLimitData(TypedDict):
    rate: int
    per: float
    bucket: BucketType
    priority: int
    exempt: ExemptCallable | None



"""Copyright 2024 PythonistaGuild

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from __future__ import annotations

import base64
import datetime
import hashlib
import json
import logging
import secrets
from typing import TYPE_CHECKING, Any

import itsdangerous
import redis.asyncio as redis
from starlette.datastructures import MutableHeaders
from starlette.requests import HTTPConnection

from ..redis import Redis


if TYPE_CHECKING:
    import redis.asyncio as redis
    from starlette.types import ASGIApp, Message, Receive, Scope, Send

    from ..redis import Redis


logger: logging.Logger = logging.getLogger(__name__)


class Storage:
    __slots__ = "pool"

    def __init__(self, *, redis: Redis) -> None:
        self.pool: redis.Redis = redis.pool

    async def get(self, data: dict[str, Any]) -> dict[str, Any]:
        expiry: datetime.datetime = datetime.datetime.fromisoformat(data["expiry"])
        key: str = data["_session_secret_key"]

        if expiry <= datetime.datetime.now():
            await self.pool.delete(key)  # type: ignore
            return {}

        session: Any = await self.pool.get(key)  # type: ignore
        return json.loads(session) if session else {}

    async def set(self, key: str, value: dict[str, Any], *, max_age: int) -> None:
        await self.pool.set(key, json.dumps(value), ex=max_age)  # type: ignore

    async def delete(self, key: str) -> None:
        await self.pool.delete(key)  # type: ignore


class SessionMiddleware:
    def __init__(
        self,
        app: ASGIApp,
        *,
        name: str | None = None,
        secret: str | None = None,
        max_age: int | None = None,
        same_site: str = "lax",
        secure: bool = True,
        redis: Redis,
    ) -> None:
        self.app: ASGIApp = app
        self.name: str = name or "__session_cookie"
        self.secret: str = secret or secrets.token_urlsafe(
            128
        )  # set this if you don't want to invalidate sessions on restart
        self.max_age: int = max_age or (60 * 60 * 24 * 7)  # 7 days; 1 week
        self.signing: itsdangerous.Signer = itsdangerous.Signer(self.secret, digest_method=hashlib.sha256)
        self.storage: Storage = Storage(redis=redis)

        self.flags: str = f"HttpOnly; SameSite={same_site}; Path=/{'; secure' if secure else ''}"

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] not in ("http", "websocket"):
            await self.app(scope, receive, send)
            return

        # Use this to cover both websocket connections and http connections
        connection: HTTPConnection = HTTPConnection(scope, receive)
        session: dict[str, Any]

        try:
            cookie: bytes = connection.cookies[self.name].encode("utf-8")
        except KeyError:
            cookie = b""

        try:
            unsigned: str = self.signing.unsign(base64.b64decode(cookie)).decode("utf-8")
            data: dict[str, Any] = json.loads(unsigned)
            session = await self.storage.get(data)
        except (KeyError, itsdangerous.BadSignature):
            session = {}

        original: dict[str, Any] = session.copy()
        scope["session"] = session

        async def wrapper(message: Message) -> None:
            if message["type"] != "http.response.start":
                await send(message)
                return

            headers: MutableHeaders = MutableHeaders(scope=message)
            secret_key: str = scope["session"].get("_session_secret_key", secrets.token_urlsafe(64))

            # At this point we can assume that the server has cleared the session...
            if not scope["session"] and original:
                await self.storage.delete(original["_session_secret_key"])
                headers.append("Set-Cookie", self.cookies(value="null", clear=True))

            # Server has updated the session data so we need to set a new cookie...
            elif scope["session"] != original:
                expiry = datetime.datetime.now() + datetime.timedelta(seconds=self.max_age)
                scope["session"]["_session_secret_key"] = secret_key

                cookie_: dict[str, str] = {"_session_secret_key": secret_key, "expiry": expiry.isoformat()}
                signed: bytes = base64.b64encode(self.signing.sign(json.dumps(cookie_)))
                headers.append("Set-Cookie", self.cookies(value=signed.decode("utf-8")))

                await self.storage.set(secret_key, scope["session"], max_age=self.max_age)

            elif not session and not original and cookie:
                headers.append("Set-Cookie", self.cookies(value="null", clear=True))

            await send(message)

        await self.app(scope, receive, wrapper)

    def cookies(self, *, value: str, clear: bool = False) -> str:
        if clear:
            return f"{self.name}={value}; {self.flags}; Max-Age=0"

        return f"{self.name}={value}; {self.flags}; Max-Age={self.max_age}"



"""Copyright 2024 PythonistaGuild

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from __future__ import annotations

import asyncio
import inspect
import logging
from collections.abc import Callable, Iterator
from typing import TYPE_CHECKING, Any, Self, TypeAlias, TypedDict, Unpack

from starlette.applications import Starlette
from starlette.requests import Request
from starlette.responses import Response
from starlette.routing import Route, WebSocketRoute
from starlette.types import Receive, Scope, Send
from starlette.websockets import WebSocket

from .types_.core import RouteCoro


if TYPE_CHECKING:
    from starlette.middleware import Middleware
    from starlette.types import Message, Receive, Scope, Send

    from .types_.core import Methods, RouteOptions
    from .types_.limiter import BucketType, ExemptCallable, RateLimitData


access_logger: logging.Logger = logging.getLogger("Route")


class ApplicationOptions(TypedDict, total=False):
    prefix: str
    views: list[View]
    access_log: bool
    middleware: list[Middleware]


__all__ = ("Application", "View", "route", "limit")


class _Route:
    def __init__(self, **kwargs: Unpack[RouteOptions]) -> None:
        self._path: str = kwargs["path"]
        self._coro: Callable[..., RouteCoro] = kwargs["coro"]
        self._methods: Methods = kwargs["methods"]
        self._prefix: bool = kwargs["prefix"]
        self._limits: list[RateLimitData] = kwargs.get("limits", [])
        self._is_websocket: bool = kwargs.get("websocket", False)
        self._view: View | None = None

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> Response | None:
        request: Request | WebSocket = (
            WebSocket(scope, receive, send) if scope["type"] == "websocket" else Request(scope, receive, send)
        )

        response: Response | None = await self._coro(self._view, request)
        if response is None:
            response = Response(status_code=500, content="Internal Server Error")

        await response(scope, receive, send)


LimitDecorator: TypeAlias = Callable[..., RouteCoro] | _Route
T_LimitDecorator: TypeAlias = Callable[..., LimitDecorator]


def route(
    path: str,
    /,
    *,
    methods: Methods = ["GET"],
    prefix: bool = True,
    websocket: bool = False,
) -> Callable[..., _Route]:
    def decorator(coro: Callable[..., RouteCoro]) -> _Route:
        if not asyncio.iscoroutinefunction(coro):
            raise RuntimeError("Route callback must be a coroutine function.")

        disallowed: list[str] = ["GET", "POST", "PUT", "DELETE", "PATCH", "HEAD", "OPTIONS"]
        if coro.__name__.upper() in disallowed:
            raise ValueError(f"Route callback function must not be named any: {', '.join(disallowed)}")

        limits: list[RateLimitData] = getattr(coro, "__limits__", [])
        return _Route(path=path, coro=coro, methods=methods, prefix=prefix, limits=limits, websocket=websocket)

    return decorator


def limit(
    rate: int,
    per: float,
    *,
    bucket: BucketType = "ip",
    priority: int = 0,
    exempt: ExemptCallable | None = None,
) -> T_LimitDecorator:
    def decorator(coro: Callable[..., RouteCoro] | _Route) -> LimitDecorator:
        limits: RateLimitData = {"rate": rate, "per": per, "bucket": bucket, "priority": priority, "exempt": exempt}

        if isinstance(coro, _Route):
            coro._limits.append(limits)
        else:
            setattr(coro, "__limits__", [limits])

        return coro

    return decorator


class Application(Starlette):
    __routes__: list[Route | WebSocketRoute]

    def __init__(self, *args: Any, **kwargs: Unpack[ApplicationOptions]) -> None:
        self._views: list[View] = []

        self._prefix: str = kwargs.pop("prefix", "")
        self._access_log: bool = kwargs.pop("access_log", True)
        views: list[View] = kwargs.pop("views", [])

        super().__init__(*args, **kwargs)  # type: ignore

        self.add_view(self)
        for view in views:
            self.add_view(view)

    def __new__(cls, *args: Any, **kwargs: Any) -> Self:
        self: Self = super().__new__(cls)
        self.__routes__ = []

        name: str = cls.__name__

        for _, member in inspect.getmembers(self, predicate=lambda m: isinstance(m, _Route)):
            member._view = self
            path: str = member._path

            for method in member._methods:
                method = method.lower()

                # Due to the way Starlette works, this allows us to have schema documentation...
                setattr(member, method, member._coro)

            new: WebSocketRoute | Route

            if member._is_websocket:
                new = WebSocketRoute(path=path, endpoint=member, name=f"{name}.{member._coro.__name__}")
            else:
                new = Route(path=path, endpoint=member, methods=member._methods, name=f"{name}.{member._coro.__name__}")

            new.limits = getattr(member, "_limits", [])  # type: ignore
            self.__routes__.append(new)

        return self

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] != "http" or not self._access_log:
            return await super().__call__(scope, receive, send)

        method: str = scope["method"]
        path: str = scope["path"]
        client: str = f"{scope['client'][0]}:{scope['client'][1]}"
        version: str = scope["http_version"]

        async def inspect_response(message: Message) -> None:
            nonlocal method, path, client

            if message["type"] == "http.response.start":
                status_code: int = message.get("status", 200)
                msg: str = f'{client} - "{method} {path} HTTP/{version}" '

                access_logger.info(msg, extra={"status": status_code})

            await send(message)

        await super().__call__(scope, receive, inspect_response)

    @property
    def prefix(self) -> str:
        return self._prefix

    @property
    def views(self) -> list[View]:
        return self._views

    def add_view(self, view: View | Self) -> None:
        if view in self._views:
            msg: str = f"A view with the name '{view.name}' has already been added to this application."
            raise RuntimeError(msg)

        routes: list[Route | WebSocketRoute] = getattr(view, "__routes__", [])

        for route_ in routes:
            path = f"/{self._prefix.lstrip('/')}{route_.path}" if self._prefix else route_.path

            if isinstance(route_, WebSocketRoute):
                new = WebSocketRoute(path, endpoint=route_.endpoint, name=route_.name)
            else:
                methods: list[str] | None = list(route_.methods) if route_.methods else None
                new = Route(path, endpoint=route_.endpoint, methods=methods, name=route_.name)

            new.limits = route_.limits  # type: ignore
            self.router.routes.append(new)

        if isinstance(view, View):
            self._views.append(view)


class View:
    __routes__: list[Route | WebSocketRoute]

    def __new__(cls, *args: Any, **kwargs: Any) -> Self:
        self = super().__new__(cls)
        name = cls.__name__

        self.__routes__ = []

        for _, member in inspect.getmembers(self, predicate=lambda m: isinstance(m, _Route)):
            member._view = self
            path: str = member._path

            if member._prefix:
                path = f"/{name.lower()}/{path.lstrip('/')}"

            for method in member._methods:
                method = method.lower()

                # Due to the way Starlette works, this allows us to have schema documentation...
                setattr(member, method, member._coro)

            new: WebSocketRoute | Route

            if member._is_websocket:
                new = WebSocketRoute(path=path, endpoint=member, name=f"{name}.{member._coro.__name__}")
            else:
                new = Route(path=path, endpoint=member, methods=member._methods, name=f"{name}.{member._coro.__name__}")

            new.limits = getattr(member, "_limits", [])  # type: ignore
            self.__routes__.append(new)

        return self

    @property
    def name(self) -> str:
        return self.__class__.__name__.lower()

    def __repr__(self) -> str:
        return f"View: name={self.__class__.__name__}, routes={self.__routes__}"

    def __getitem__(self, index: int) -> Route | WebSocketRoute:
        return self.__routes__[index]

    def __len__(self) -> int:
        return len(self.__routes__)

    def __iter__(self) -> Iterator[Route | WebSocketRoute]:
        return iter(self.__routes__)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, View):
            return False

        return self.name == other.name



"""
The MIT License (MIT)

Copyright (c) 2015-present Rapptz
Copyright (c) 2024 PythonistaGuild

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

import logging
import os
import pathlib
import sys
from typing import Any


__all__ = ("setup_logging",)


def is_docker() -> bool:
    path: pathlib.Path = pathlib.Path("/proc/self/cgroup")

    exists: bool = path.exists()
    is_file: bool = path.is_file()
    return exists or (is_file and any("docker" in line for line in path.open()))


def stream_supports_colour(stream: Any) -> bool:
    is_a_tty = hasattr(stream, "isatty") and stream.isatty()

    # Pycharm and Vscode support colour in their inbuilt editors
    if "PYCHARM_HOSTED" in os.environ or os.environ.get("TERM_PROGRAM") == "vscode":
        return is_a_tty

    if sys.platform != "win32":
        # Docker does not consistently have a tty attached to it
        return is_a_tty or is_docker()

    # ANSICON checks for things like ConEmu
    # WT_SESSION checks if this is Windows Terminal
    return is_a_tty and ("ANSICON" in os.environ or "WT_SESSION" in os.environ)


def stream_supports_rgb(stream: Any) -> bool:
    if not stream_supports_colour(stream):
        return False

    if "COLORTERM" in os.environ:
        return os.environ["COLORTERM"] in ("truecolor", "24bit")

    return False


class ColourFormatter(logging.Formatter):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

        self._handler: logging.StreamHandler[Any] = kwargs.get("handler", logging.StreamHandler())

        self._supports_colour: bool = stream_supports_colour(self._handler.stream)
        self._supports_rgb: bool = stream_supports_rgb(self._handler.stream)

        self._colours: dict[int, str] = {}
        self._RESET: str = "\033[0m"

        self.green: str
        self.yellow: str
        self.red: str

        if self._supports_rgb:
            self._colours = {
                logging.DEBUG: "\x1b[40;1m",
                logging.INFO: "\x1b[38;2;100;55;215;1m",
                logging.WARNING: "\x1b[38;2;204;189;51;1m",
                logging.ERROR: "\x1b[38;2;161;38;46m",
                logging.CRITICAL: "\x1b[48;2;161;38;46",
            }

            self.green = "\x1b[38;2;12;218;120;1m"
            self.yellow = "\x1b[38;2;204;189;51;1m"
            self.red = "\x1b[38;2;161;38;46m"

        elif self._supports_colour:
            self._colours = {
                logging.DEBUG: "\x1b[40;1m",
                logging.INFO: "\x1b[34;1m",
                logging.WARNING: "\x1b[33;1m",
                logging.ERROR: "\x1b[31m",
                logging.CRITICAL: "\x1b[41",
            }

            self.green = "\x1b[32m"
            self.yellow = "\x1b[33m"
            self.red = "\x1b[31m"

        self._FORMATS: dict[int, logging.Formatter] = {
            level: logging.Formatter(
                f"\x1b[30;1m%(asctime)s\x1b[0m {colour}%(levelname)-8s\x1b[0m {colour}%(name)s\x1b[0m %(message)s"
            )
            for level, colour in self._colours.items()
        }

    def format(self, record: logging.LogRecord) -> str:
        formatter: logging.Formatter | None = self._FORMATS.get(record.levelno, None)
        if formatter is None:
            formatter = self._FORMATS[logging.DEBUG]

        # Override the traceback to always print in red
        if record.exc_info:
            text = formatter.formatException(record.exc_info)
            record.exc_text = f"\x1b[31m{text}\x1b[0m"

        output = formatter.format(record)

        if record.name == "Route":
            status: int | None = record.__dict__.get("status", None)
            if not status:
                record.exc_text = None
                return output

            if status >= 400:
                output = f"{output}{self._RESET} {self.red}{status}{self._RESET}"
            elif status >= 300:
                output = f"{output}{self._RESET} {self.yellow}{status}{self._RESET}"
            else:
                output = f"{output}{self._RESET} {self.green}{status}{self._RESET}"

        # Remove the cache layer
        record.exc_text = None
        return output


ColorFormatter = ColourFormatter


def setup_logging(
    *,
    handler: logging.Handler | None = None,
    formatter: logging.Formatter | None = None,
    level: int | None = None,
    root: bool = True,
) -> None:
    if level is None:
        level = logging.INFO

    if handler is None:
        handler = logging.StreamHandler()

    if formatter is None:
        if isinstance(handler, logging.StreamHandler) and stream_supports_colour(handler.stream):  # type: ignore
            formatter = ColourFormatter()
        else:
            dt_fmt = "%Y-%m-%d %H:%M:%S"
            formatter = logging.Formatter("[{asctime}] [{levelname:<8}] {name}: {message}", dt_fmt, style="{")

    if root:
        logger = logging.getLogger()
    else:
        library, _, _ = __name__.partition(".")
        logger = logging.getLogger(library)

    handler.setFormatter(formatter)
    logger.setLevel(level)
    logger.addHandler(handler)  # type: ignore



"""Copyright 2024 PythonistaGuild

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from starlette.requests import Request
from starlette.responses import JSONResponse, Response
from starlette.routing import Route
from starlette.types import ASGIApp, Receive, Scope, Send

from ..limiter import RateLimit, Store
from ..redis import Redis


if TYPE_CHECKING:
    from starlette.routing import Route
    from starlette.types import ASGIApp, Receive, Scope, Send

    from ..redis import Redis
    from ..types_.limiter import BucketType, ExemptCallable, RateLimitData, ResponseCallback


logger: logging.Logger = logging.getLogger(__name__)


class RatelimitMiddleware:
    def __init__(
        self,
        app: ASGIApp,
        *,
        ignore_localhost: bool = True,
        global_limits: list[RateLimitData] = [],
        response_callback: ResponseCallback | None = None,
        redis: Redis | None = None,
    ) -> None:
        self.app: ASGIApp = app

        self._ignore_local: bool = ignore_localhost
        self._global_limits: list[RateLimitData] = global_limits

        self._store: Store = Store(redis=redis)
        self._response_callback: ResponseCallback = response_callback or self.default_response

    async def default_response(self, request: Request, retry: float) -> Response:
        return JSONResponse(
            {"error": "You are requesting too fast."},
            status_code=429,
            headers={"Retry-After": str(retry)},
        )

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] != "http":
            return await self.app(scope, receive, send)

        request: Request = Request(scope)
        forwarded: str | None = request.headers.get("X-Forwarded-For", None)

        routes: list[Route] = scope["app"].routes
        route: Route | None = None

        for r in routes:
            methods: set[str] = r.methods or set()
            if r.path == request.url.path and request.method in methods:
                route = r
                break

        route_limits: list[RateLimitData] = sorted(getattr(route, "limits", []), key=lambda x: x["priority"])

        for limit in self._global_limits + route_limits:
            is_exempt: bool = False
            exempt: ExemptCallable | None = limit["exempt"]

            if exempt is not None:
                is_exempt: bool = await exempt(request)

            if is_exempt:
                continue

            bucket: BucketType = limit["bucket"]
            if bucket == "ip":
                if not request.client and not forwarded:
                    logger.warning("Could not determine the IP address while ratelimiting! Ignoring...")
                    return await self.app(scope, receive, send)

                # forwarded or client.host will exist at this point...
                key: str = forwarded.split(",")[0] if forwarded else request.client.host  # type: ignore

                if self._ignore_local and key in ("127.0.0.1", "::1", "localhost", "0.0.0.0"):
                    return await self.app(scope, receive, send)
            else:
                key: str | None = await bucket(request)
                if not key:
                    # Request is assumed exempt from ratelimiting...
                    return await self.app(scope, receive, send)

            encapsulated: RateLimit = RateLimit(limit["rate"], limit["per"])
            if retry := await self._store.update(key, encapsulated):
                response: Response = await self._response_callback(request, retry)
                await response(scope, receive, send)
                return

        await self.app(scope, receive, send)



"""Copyright 2024 PythonistaGuild

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from __future__ import annotations

import datetime
import logging
from typing import TYPE_CHECKING, TypedDict


if TYPE_CHECKING:
    from .redis import Redis


logger: logging.Logger = logging.getLogger(__name__)


class TatStore(TypedDict):
    tat: datetime.datetime
    limit: RateLimit


class RateLimit:
    def __init__(self, rate: int, per: float) -> None:
        self.rate: int = rate
        self.period: datetime.timedelta = datetime.timedelta(seconds=per)

    @property
    def inverse(self) -> float:
        return self.period.total_seconds() / self.rate


class Store:
    def __init__(self, redis: Redis | None = None) -> None:
        self.redis: Redis | None = redis
        self._keys: dict[str, TatStore] = {}

    async def get_tat(self, key: str, /) -> datetime.datetime:
        now: datetime.datetime = datetime.datetime.now(tz=datetime.UTC)

        if self.redis:
            value: str | None = await self.redis.pool.get(key)  # type: ignore
            return datetime.datetime.fromisoformat(value) if value else now  # type: ignore

        return self._keys.get(key, {"tat": now}).get("tat", now)

    async def set_tat(self, key: str, /, *, tat: datetime.datetime, limit: RateLimit) -> None:
        if self.redis:
            await self.redis.pool.set(key, tat.isoformat(), ex=limit.period.total_seconds() + 60)  # type: ignore
        else:
            self._keys[key] = {"tat": tat, "limit": limit}

    async def update(self, key: str, limit: RateLimit) -> bool | float:
        now: datetime.datetime = datetime.datetime.now(tz=datetime.UTC)
        tat: datetime.datetime = max(await self.get_tat(key), now)

        # Clear stale keys...
        for ek, ev in self._keys.copy().items():
            if (now - ev["tat"]).total_seconds() > ev["limit"].period.total_seconds() + 60:
                del self._keys[ek]

        separation: float = (tat - now).total_seconds()
        max_interval: float = limit.period.total_seconds() - limit.inverse

        if separation > max_interval:
            return separation - max_interval

        new_tat: datetime.datetime = max(tat, now) + datetime.timedelta(seconds=limit.inverse)
        await self.set_tat(key, tat=new_tat, limit=limit)

        return False
